
#include "MovieReader.h"


extern void Ding(char *msg, HRESULT hr);

#define DINGRET(x) if FAILED(hr) { Ding(x, hr); return(hr); }

#define TICKS_PER_SEC ((double)(1.0E+7))


MovieReader::MovieReader() {
	mSyncReader = NULL;
	VideoStreamNum = 0;
	Compressed = FALSE;
	mProfile = NULL;
	mProps = NULL;
	mMedia = NULL;
	mVideo = NULL;
	mSample = NULL;
}

MovieReader::~MovieReader() {
	SAFE_RELEASE(mProfile);
	SAFE_RELEASE(mProps);
	SAFE_RELEASE(mSyncReader);
	SAFE_RELEASE(mSample);
	if (mMedia != NULL) delete[] mMedia;
}


HRESULT MovieReader::Open(const WCHAR *filename) {
	HRESULT	hr=S_OK;

	if (mSyncReader == NULL) {
		hr = WMCreateSyncReader(NULL, 0, &mSyncReader);
		DINGRET("Could not create SyncReader");
	}

	hr = mSyncReader->Open(filename);
	DINGRET("Could not open input file");

	hr = mSyncReader->QueryInterface(IID_IWMProfile, (VOID **)&mProfile);
	DINGRET("Could not query interface for IWMProfile");

	hr = GetStreamNumbers();
	DINGRET("Could not get stream numbers");

	WMT_STREAM_SELECTION	wmtSS = WMT_ON;
	hr = mSyncReader->SetStreamsSelected(1, &VideoStreamNum, &wmtSS);
	DINGRET("Could not select video stream");

	mSyncReader->SetReadStreamSamples(VideoStreamNum, Compressed);
	DINGRET("Could not set compression of video stream");

	FrameNum = 0;
	FrameTime = (double)0.0;
	FramesPerSec = (double)1.0;
	FrameDuration = (double)0.0;
	FrameOutputNum = 0;
	FrameFlags = 0;
	TimeSyncd = TRUE;

	return(hr);
}

HRESULT MovieReader::Close(void) {
	HRESULT	hr=S_OK;

	if (mSyncReader != NULL) {
		hr = mSyncReader->Close();
		DINGRET("Could not close SyncReader");
	}
	SAFE_RELEASE(mProfile);
	SAFE_RELEASE(mProps);
	SAFE_RELEASE(mSyncReader);
	if (mMedia != NULL) delete[] mMedia;

	VideoStreamNum = 0;
	Compressed = FALSE;
	mVideo = NULL;

	return(hr);
}


HRESULT MovieReader::GetStreamNumbers(void) {
	HRESULT			hr = S_OK;
	IWMStreamConfig	*pStream = NULL;
	DWORD			dwStreams = 0;
	GUID			pguidStreamType;

	hr = mProfile->GetStreamCount(&dwStreams);
	DINGRET("Could not get stream count");

	VideoStreamNum = 0;

	for (DWORD i=0; i<dwStreams; i++) {
		hr = mProfile->GetStream(i, &pStream);
		DINGRET("Could not get an individual stream");

		hr = pStream->GetStreamType(&pguidStreamType);
		DINGRET("Could not get an individual stream type");

		if (pguidStreamType == WMMEDIATYPE_Video) {
			hr = pStream->GetStreamNumber(&VideoStreamNum);
			DINGRET("Could not get video stream number");

			DWORD	cbType = 0;

			hr = pStream->QueryInterface(IID_IWMMediaProps, (VOID **)&mProps);
			DINGRET("Could not query interface for Media Properties");

			hr = mProps->GetMediaType(NULL, &cbType);
			DINGRET("Could not get media type length");

			if (mMedia != NULL) delete[] mMedia;
			mMedia = (WM_MEDIA_TYPE *)new BYTE[cbType];

			hr = mProps->GetMediaType(mMedia, &cbType);
			DINGRET("Could not get media type structure");

			mVideo = (WMVIDEOINFOHEADER *)mMedia->pbFormat;
		}
	}

	return(hr);
}


HRESULT MovieReader::SetInfile(CString &filename) {
	WCHAR	filenameW[1000];
	LPCSTR	filenameC = filename;

	mbstowcs(filenameW, filenameC, strlen(filenameC)+1);
	Close();
	return(Open(filenameW));
}


RECT MovieReader::GetFrameRect(void) {
	RECT	rect = {0, 0, 0, 0};

	if (mVideo != NULL) {
		rect = mVideo->rcSource;
	}
	return(rect);
}


int MovieReader::NextFrame(void) {
	FrameNum++;
	TimeSyncd = FALSE;
	return(FrameNum);
}

int MovieReader::PrevFrame(void) {
	FrameNum--;
	if (FrameNum <= 0) FrameNum = 1;
	TimeSyncd = FALSE;
	return(FrameNum);
}

int MovieReader::SetFrameNum(int newnum) {
	FrameNum = newnum;
	if (FrameNum <= 0) FrameNum = 1;
	TimeSyncd = FALSE;
	return(FrameNum);
}

// Does not actually set by time, but by a guess at the frame, because we can
// determine time from the interface but not frame
int MovieReader::SetFrameTime(double newtime) {
	FrameNum = (int)(newtime * FramesPerSec) + 1;
	if (FrameNum <= 0) FrameNum = 1;
	TimeSyncd = FALSE;
	return(FrameNum);
}


int MovieReader::GetFrameNum(void) {
	if (!TimeSyncd) ReadFrame();
	return(FrameNum);
}

double MovieReader::GetFrameTime(void) {
	if (!TimeSyncd) ReadFrame();
	return(FrameTime);
}

double MovieReader::GetFramesPerSec(void) {
	return(FramesPerSec);
}


HRESULT MovieReader::ReadFrame(void) {
	HRESULT	hr = S_OK;
	QWORD	temptime, tempdur;
	WORD	tempstream;
	double	lasttime;

//	hr = mSyncReader->SetRangeByFrame(VideoStreamNum, FrameNum, 0);		#### Don't know how to do this yet
	DINGRET("Could not set frame number to read");

	hr = mSyncReader->GetNextSample(VideoStreamNum, &mSample, &temptime, &tempdur, &FrameFlags, &FrameOutputNum, &tempstream);
	DINGRET("Could not get video sample");

	lasttime = FrameTime;
	FrameTime = (double)(__int64)temptime / TICKS_PER_SEC;		// Dopey compiler doesn't convert from unsigned __int64 (QWORD) to double
	FramesPerSec = 1.0 / (FrameTime - lasttime);				// #### CHANGE will get confused when jumping around
	FrameDuration = (double)(__int64)tempdur / TICKS_PER_SEC;	// Dopey compiler doesn't convert from unsigned __int64 (QWORD) to double
	TimeSyncd = TRUE;

	return(hr);
}

HRESULT MovieReader::GetFrame(INSSBuffer **outSample, BITMAPINFO **outInfo) {
	HRESULT	hr = S_OK;

	if (!TimeSyncd) hr=ReadFrame();

	*outSample = mSample;

	*outInfo = (BITMAPINFO *)&mVideo->bmiHeader;	// Assumes no color map!

	return(hr);
}


